import torch
from torch.autograd        import grad
from sparse_distance_bmp   import sparse_distance_bmp
from pykeops.torch         import Kernel

import numpy as np
from matplotlib import pyplot as plt
from scipy import misc
from time  import time

use_cuda = torch.cuda.is_available()
tensor   = torch.cuda.FloatTensor if use_cuda else torch.FloatTensor 

plt.ion()
plt.show()

s2v = lambda x : tensor([x])


from scipy.ndimage.filters import gaussian_filter

def LoadImage(fname) :
    img = misc.imread(fname, flatten = True) # Grayscale
    #img = gaussian_filter(img, 1, mode='nearest')
    img = (img[::-1, :])  / 255.
    img = np.swapaxes(img, 0,1 )
    return tensor( 1 - img )

# Load the pngs
dataset = "bars"
datasets = {
    "bars"     : ("data/hor_bars_1.png", "data/hor_bars_2.png"),
    "worms_c"  : ("data/worm_c.png",     "data/worm_d.png"),
}
if dataset == "knees_2D" :
    sample = 8
    smooth = True # As the structure is very thin, sampling on a grid every 8 pixels looks weird...
                  # we smooth the gradient a little bit to prevent this visualization artifact from being visible
    scale_grad = 0.04
else :
    sample = 4
    smooth = False
    scale_grad = 0.02

source = LoadImage(datasets[dataset][0])
target = LoadImage(datasets[dataset][1])

# The images are rescaled to fit into the unit square ==========================================
scale = source.shape[0]
affine = tensor( [[ 1, 0, 0 ],
                  [ 0, 1, 0 ] ]) / scale

# Parameters of our data attachment term =======================================================
experiments = {}

experiments["dataset"] = {
    "formula"     : "kernel",
    "id"          : Kernel( "-distance(x,y)" ),
    "gamma"       : s2v( 1. ),
}
experiments["softmin_med"] = {
    "formula"        : "hausdorff",
    "cost"           : "bregman",
    "distance_field" : "sinkhorn",
    "nits"           : 1,
    "id"          : Kernel( "gaussian(x,y)" ),
    "epsilon"     : s2v(   .1**2),
    "gamma"       : s2v( 1/.1**2),
}
experiments["softmin_med_a"] = {
    "formula"        : "hausdorff",
    "cost"           : "bregman_a",
    "distance_field" : "sinkhorn",
    "nits"           : 1,
    "id"          : Kernel( "gaussian(x,y)" ),
    "epsilon"     : s2v(   .1**2),
    "gamma"       : s2v( 1/.1**2),
}
experiments["softmin_med_b"] = {
    "formula"        : "hausdorff",
    "cost"           : "bregman_b",
    "distance_field" : "sinkhorn",
    "nits"           : 1,
    "id"          : Kernel( "gaussian(x,y)" ),
    "epsilon"     : s2v(   .1**2),
    "gamma"       : s2v( 1/.1**2),
}


experiments["dataset_2"] = {
    "formula"     : "kernel",
    "id"          : Kernel( "-distance(x,y)" ),
    "gamma"       : s2v( 1. ),
}


# We'll save the output wrt. the number of iterations
display = True
plt.figure(figsize=(10,10))


def test(name, params, verbose=True) :
    params["kernel_heatmap_range"] = (0,1,100)

    # Compute the cost and gradient ============================================================
    t_0 = time()
    cost, grad_src, heatmaps = sparse_distance_bmp(params, source, target, 
                                                           affine, affine, 
                                                           normalize=True, info=display )
    t_1 = time()
    if verbose : print("{} : {:.2f}s, cost = {:.6f}".format( name, t_1-t_0, cost.item()) )

    # Display ==================================================================================
    grad_src = - grad_src # N.B.: We want to visualize a descent direction, not the opposite!
    if display :
        plt.clf()

        # Source + Target :
        source_plot = .3*source.cpu().numpy()
        target_plot = .3*target.cpu().numpy()
        img_plot    = np.dstack( ( np.ones(source_plot.shape) - target_plot, 
                                   np.ones(source_plot.shape) - source_plot - target_plot, 
                                   np.ones(source_plot.shape) - source_plot ) )
        
        plt.imshow( np.swapaxes(img_plot,0,1), origin="lower", extent=(0,1,0,1))

        if name not in  ["dataset", "dataset_2"] :
            # Subsample the gradient field :
            grad_plot = grad_src.cpu().numpy()
            grad_plot = grad_plot[::sample, ::sample, :]
            if smooth :
                grad_plot = gaussian_filter(grad_plot, [1,1,0], mode='nearest')

            X,Y   = np.meshgrid( np.linspace(0, 1, grad_plot.shape[0]+1)[:-1] + .5/(sample*grad_plot.shape[0]), 
                                np.linspace(0, 1, grad_plot.shape[1]+1)[:-1] + .5/(sample*grad_plot.shape[1]) )
            U, V  = grad_plot[:,:,0], grad_plot[:,:,1]
            norms = np.sqrt( (grad_plot[:,:,1]**2 + grad_plot[:,:,0]**2) )
            mask  = (norms>0)
            scale = np.mean( norms[mask] )

            if name == "softmin_sml" : 
                # At this point in the slides, I don't want to explain the projection
                # artifact so I cheat a little bit
                mask = mask * (norms < 1.5*scale)


            if smooth : mask  = (source_plot[::sample, ::sample]>0)

            Y, X, U, V = Y[mask], X[mask], U[mask], V[mask]
            plt.quiver( Y, X, U, V, 
                        scale = scale_grad*scale, scale_units="dots", color="#5CBF3A", zorder=3, width=0.0025)

            # "Distance" fields :
            if heatmaps is not None :
                heatmaps.plot(plt.gca(), {})
            
        # Save result in the "output/" folder :
        plt.savefig("output/projection/" + dataset + "_" + name +".png") 
        plt.pause(.01)


#test("hausdorff_KL_1_1", experiments["hausdorff_KL_1_1"], verbose=False) # one run for nothing
for name, params in experiments.items() :
    test(name, params)

print("Done.")
plt.show(block=True)
